import requests
import json
from collections import namedtuple
from requests.auth import HTTPBasicAuth
from requests_toolbelt.utils import dump


class DomoAPITransport:
    """Essentially a wrapper around the 'requests' library to make
    it easier to interact with the Domo API.

    OAuth2 authentication is handled automatically, as well as the
    serialization and deserialization of objects.
    """

    def __init__(self, client_id, client_secret, api_host, use_https, logger):
        self.apiHost = self._build_apihost(api_host, use_https)
        self.clientId = client_id
        self.clientSecret = client_secret
        self.logger = logger
        self._renew_access_token()

    @staticmethod
    def _build_apihost(host, use_https):
        if use_https:
            host = 'https://' + host
        else:
            host = 'http://' + host
        return host

    def get(self, url, params):
        headers = self._headers_default_receive_json()
        return self.request(url, HTTPMethod.GET, headers, params)

    def get_csv(self, url, params):
        headers = self._headers_receive_csv()
        return self.request(url, HTTPMethod.GET, headers, params)

    def post(self, url, body, params):
        headers = self._headers_send_json()
        return self.request(url, HTTPMethod.POST, headers, params,
                            self._obj_to_json(body))

    def put(self, url, body):
        headers = self._headers_send_json()
        return self.request(url, HTTPMethod.PUT, headers, {},
                            self._obj_to_json(body))

    def put_csv(self, url, body):
        headers = self._headers_send_csv()
        return self.request(url, HTTPMethod.PUT, headers, {}, body)

    def put_gzip(self, url, body):
        headers = self._headers_send_gzip()
        return self.request(url, HTTPMethod.PUT, headers, {}, body)

    def patch(self, url, body):
        headers = self._headers_send_json()
        return self.request(url, HTTPMethod.PATCH, headers, {},
                            self._obj_to_json(body))

    def delete(self, url):
        headers = self._headers_default_receive_json()
        return self.request(url, HTTPMethod.DELETE, headers)

    def request(self, url, method, headers, params=None, body=None):
        url = self.apiHost + url
        self.logger.debug('{} {} {}'.format(method, url, body))
        request_args = {'method': method, 'url': url, 'headers': headers,
                        'params': params, 'data': body, 'stream': True}

        response = requests.request(**request_args)
        if response.status_code == requests.codes.UNAUTHORIZED:
            self._renew_access_token()
            headers['Authorization'] = 'bearer ' + self.access_token
            response = requests.request(**request_args)
        return response

    def _renew_access_token(self):
        self.logger.debug("Renewing Access Token")
        url = self.apiHost + '/oauth/token?grant_type=client_credentials'
        response = requests.post(url=url, auth=HTTPBasicAuth(self.clientId, self.clientSecret))
        if response.status_code == requests.codes.OK:
            self.access_token = response.json()['access_token']
        else:
            self.logger.debug('Error retrieving access token: ' + self.dump_response(response))
            raise Exception("Error retrieving a Domo API Access Token: " + response.text)

    def dump_response(self, response):
        data = dump.dump_all(response)
        return str(data.decode('utf-8'))

    @staticmethod
    def _obj_to_json(obj):
        return json.dumps(obj, default=str)

    def _headers_default_receive_json(self):
        return {
            'Authorization': 'bearer ' + self.access_token,
            'Accept': 'application/json'
        }

    def _headers_send_json(self):
        headers = self._headers_default_receive_json()
        headers['Content-Type'] = 'application/json'
        return headers

    def _headers_send_csv(self):
        headers = self._headers_default_receive_json()
        headers['Content-Type'] = 'text/csv'
        return headers

    def _headers_send_gzip(self):
        headers = self._headers_default_receive_json()
        headers['Content-Type'] = 'text/csv'
        headers['Content-Encoding'] = 'gzip'
        return headers

    def _headers_receive_csv(self):
        headers = self._headers_default_receive_json()
        headers['Accept'] = 'text/csv'
        return headers


class HTTPMethod:
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'



import requests

from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod

PAGE_DESC = "Page"
COLLECTION_DESC = "Collection"
URL_BASE = '/v1/pages'

PAGE_CREATION_KWARGS = [
    'parentId',
    'locked',
    'cardIds',
    'visibility'
]

PAGE_UPDATE_KWARGS = [
    'id',
    'name',
    'parentId',
    'ownerId',
    'locked',
    'collectionIds',
    'cardIds',
    'visibility'
]

COLLECTION_CREATION_KWARGS = [
    'description',
    'cardIds'
]

COLLECTION_UPDATE_KWARGS = [
    'id',
    'description',
    'cardIds',
    'title'
]


class PageClient(DomoAPIClient):
    """
        Pages
        - Programmatically manage Domo Pages
        - Docs: https://developer.domo.com/docs/data-apis/pages
    """

    def create(self, name, **kwargs):
        """Create a new page.

        >>> page = {'name':'New Page'}
        >>> new_page = domo.pages.create(**page)
        >>> print(new_page)
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}

        :Parameters:
          - `name`: The name of the new page
          - `parentId`: (optional) If present create page as subpage
          - `locked`: (optional) whether to lock the page
          - `cardIds`: (optional) cards to place on the page
          - `visibility`: (optional) dict of userIds and/or groupIds to
            give access to

        :Returns:
          - A dict representing the page
        """

        self._validate_params(kwargs, PAGE_CREATION_KWARGS)

        page_request = {'name': name}
        page_request.update(kwargs)
        return self._create(URL_BASE, page_request, {}, PAGE_DESC)

    def get(self, page_id):
        """Get a page.

        >>> page = domo.pages.get(page_id)
        >>> print(page)
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}

        :Parameters:
          - `page_id`: ID of the page to get

        :returns:
          - A dict representing the page
        """
        url = '{base}/{page_id}'.format(base=URL_BASE, page_id=page_id)
        return self._get(url, PAGE_DESC)

    def list(self, per_page=50, offset=0, limit=0):
        """List pages.
        Returns a generator that will call the API multiple times
        If limit is supplied and non-zero, returns up to limit pages

        >>> list(domo.pages.list())
        [{'id': 123456789, 'name': 'My Page', 'children': []}, ...]

        :returns:
          - A list of dicts (with nesting possible)
        """
        # API uses pagination with a max of 50 per page
        if per_page not in range(1, 51):
            raise ValueError('per_page must be between 1 and 50 (inclusive)')
        # Don't pull 50 values if user requests 10
        if limit:
            per_page = min(per_page, limit)

        params = {
            'limit': per_page,
            'offset': offset,
        }
        page_count = 0

        pages = self._list(URL_BASE, params, PAGE_DESC)
        while pages:
            for page in pages:
                yield page
                page_count += 1
                if limit and page_count >= limit:
                    return

            params['offset'] += per_page
            if limit and params['offset'] + per_page > limit:
                # Don't need to pull more than the limit
                params['limit'] = limit - params['offset']
            pages = self._list(URL_BASE, params, PAGE_DESC)

    def update(self, page_id=None, **kwargs):
        """Update a page.

        >>> print(domo.pages.get(page_id))
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}
        >>> domo.pages.update(page_id, locked=True,
                              cardIds=[54321, 13579])
        >>> print(domo.pages.get(page_id))
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': True, 'ownerId': 12345, 'cardIds': [54321, 13579],
        'visibility': {'userIds': 12345}}

        # Using **kwargs:
        >>> page = domo.pages.get(page_id)
        >>> page['cardIds'].append(new_card_id)
        >>> domo.pages.update(**page)

        :Parameters:
          - `page_id`: ID of the page to update. Can also be provided
            by supplying `id` to **kwargs. This allows for calling get,
            updating the returned object, then passing it to update.
          - `name`: (optional) rename the page
          - `parentId`: (optional) turn page into subpage, or subpage
            into top-level page if parentId is present and falsey
          - `ownerId`: (optional) change owner of the page
          - `locked`: (optional) lock or unlock the page
          - `collectionIds`: (optional) reorder collections on page
          - `cardIds`: (optional) specify which cards to have on page
          - `visibility`: (optional) change who can see the page
        """

        self._validate_params(kwargs, PAGE_UPDATE_KWARGS)
        if page_id is None and 'id' not in kwargs:
            raise TypeError("update() missing required argument: 'page_id'")
        elif 'id' in kwargs:
            if page_id is not None and page_id != kwargs['id']:
                raise ValueError('ambiguous page ID - page_id and id both '
                                 'supplied but do not match')
            page_id = kwargs['id']
            del kwargs['id']
        if not kwargs.get('parentId', True):
            # Check if parentId is present and falsey
            kwargs['parentId'] = 0

        url = '{base}/{page_id}'.format(base=URL_BASE, page_id=page_id)
        return self._update(url,
                            HTTPMethod.PUT,
                            requests.codes.NO_CONTENT,
                            kwargs,
                            PAGE_DESC)

    def delete(self, page_id):
        """Delete a page.

        :Parameters:
          - `page_id`: ID of the page to delete
        """
        url = '{base}/{page_id}'.format(base=URL_BASE, page_id=page_id)
        return self._delete(url, PAGE_DESC)

    def create_collection(self, page_id, title, **kwargs):
        """Create a collection on a page.

        >>> collection = domo.pages.create_collection(page_id,
                                                      'Collection')
        >>> print(collection)
        {'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}

        :Parameters:
          - `page_id`: ID of the page to create a collection on
          - `title`: The title of the collection
          - `description`: (optional) The description of the collection
          - `cardIds`: (optional) cards to place in the collection

        :Returns:
          - A dict representing the collection
        """

        self._validate_params(kwargs, COLLECTION_CREATION_KWARGS)
        collection_request = {'title': title}
        collection_request.update(kwargs)

        url = '{base}/{page_id}/collections'.format(
                base=URL_BASE, page_id=page_id)
        return self._create(url, collection_request, {}, COLLECTION_DESC)

    def get_collections(self, page_id):
        """Get a collections of a page

        >>> print(domo.pages.get_collections(page_id))
        [{'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}]

        :Parameters:
          - `page_id`: ID of the page

        :Returns:
          - A list of dicts representing the collections
        """
        url = '{base}/{page_id}/collections'.format(base=URL_BASE,
                                                    page_id=page_id)
        return self._get(url, COLLECTION_DESC)

    def update_collection(self, page_id, collection_id=None, **kwargs):
        """Update a collection of a page.

        >>> collections = domo.pages.get_collections(page_id)
        >>> print(collections)
        [{'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}]
        >>> collection_id = collections[0]['id']
        >>> domo.pages.update_collection(page_id, collection_id,
                                         description='description',
                                         cardIds=[54321, 13579])
        >>> print(domo.pages.get_collections(page_id))
        [{'id': 1234321, 'title': 'Collection',
        'description': 'description', 'cardIds': [54321, 13579]}]

        # Using **kwargs:
        >>> collections = domo.pages.get_collections(page_id)
        >>> collections[0]['description'] = 'Description'
        >>> domo.pages.update_collection(page_id, **collections[0])

        :Parameters:
          - `page_id`: ID of the page the collection is on
          - `collection_id`: ID of the collection. Can also be provided
            by supplying `id` to **kwargs. This allows for calling
            get_collections, updating one of the returned collections,
            then passing it to update_collection.
          - `title`: (optional) update the title
          - `description`: (optional) update the description
          - `cardIds`: (optional) update cards in the collection

        :Returns:
          - A dict representing the collection
        """

        self._validate_params(kwargs, COLLECTION_UPDATE_KWARGS)
        if collection_id is None and 'id' not in kwargs:
            raise TypeError("update() missing required argument: "
                            "'collection_id'")
        elif 'id' in kwargs:
            if collection_id is not None and collection_id != kwargs['id']:
                raise ValueError('ambiguous collection ID - collection_id and'
                                 ' id both supplied but do not match')
            collection_id = kwargs['id']
            del kwargs['id']

        url = '{base}/{page_id}/collections/{collection_id}'.format(
                base=URL_BASE, page_id=page_id, collection_id=collection_id)
        return self._update(url,
                            HTTPMethod.PUT,
                            requests.codes.NO_CONTENT,
                            kwargs,
                            COLLECTION_DESC)

    def delete_collection(self, page_id, collection_id):
        """Delete a collection from a page.

        :Parameters:
          - `page_id`: ID of the page the collection is on
          - `collection_id`: ID of the collection to delete
        """
        url = '{base}/{page_id}/collections/{collection_id}'.format(
                base=URL_BASE, page_id=page_id, collection_id=collection_id)
        return self._delete(url, COLLECTION_DESC)




import json
import math
import sys

from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.datasets import DataSetClient
from pydomo.streams import StreamClient

class UtilitiesClient(DomoAPIClient):
    def __init__(self, transport, logger):
        super(UtilitiesClient, self).__init__(transport, logger)
        self.ds = DataSetClient(self.transport, self.logger)
        self.stream = StreamClient(self.transport, self.logger)
        self.transport = transport

    def domo_schema(self, ds_id):
        this_get = self.ds.get(ds_id)
        return this_get['schema']['columns']

    def data_schema(self, df):
        col_types = dict(df.dtypes)
        output_list = []
        for key, value in col_types.items():
            output_list.append({'type': self.typeConversionText(value), 'name': key})
        return output_list

    def typeConversionText(self, dt):
        result = 'STRING'
        if dt == '<M8[ns]':
            result = 'DATETIME'
        if dt == 'int64':
            result = 'LONG'
        if dt == 'float64':
            result = 'DOUBLE'

        return result

    def identical(self, c1, c2):
        cc1 = json.dumps(c1)
        cc2 = json.dumps(c2)
        return cc1 == cc2

    def get_stream_id(self, ds_id):
        url = '/v1/streams/search?q=dataSource.id:{ds_id}'.format(ds_id=ds_id)
        all_info = self._get(url,'Search to retrieve stream id')
        return all_info[0]['id']

    def estimate_chunk_rows(self, data, kbytes=10000):
        sz = sys.getsizeof(data)
        targetSize = kbytes * 3 # compression factor
        data_rows = len(data.index)
        ch_size = data_rows
        if sz / 1000 > targetSize:
            ch_size = math.floor(data_rows*(targetSize) / (sz/1000))
        return(ch_size)

    def stream_upload(self, ds_id, df_up, warn_schema_change=True):
        domoSchema = self.domo_schema(ds_id)
        dataSchema = self.data_schema(df_up)

        stream_id = self.get_stream_id(ds_id)

        if self.identical(domoSchema,dataSchema) == False:
            new_schema = {'schema': {'columns': dataSchema}}
            url = '/v1/datasets/{ds}'.format(ds=ds_id)
            change_result = self.transport.put(url,new_schema)
            if warn_schema_change:
                print('Schema Updated')

        exec_info = self.stream.create_execution(stream_id)
        exec_id = exec_info['id']

        chunksz = self.estimate_chunk_rows(df_up)
        start = 0
        df_rows = len(df_up.index)
        end = df_rows
        if df_rows > chunksz:
            end = chunksz

        while end <= df_rows:
            df_sub = df_up.iloc[start:end, ]
            csv = df_sub.to_csv(header=False,index=False)
            self.stream.upload_part(stream_id, exec_id, start, csv)
            start = end
            end = end + chunksz

        result = self.stream.commit_execution(stream_id, exec_id)

        return result


import requests

from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod

ACCOUNT_DESC = "Account"
URL_BASE = '/v1/accounts'

ACCOUNT_CREATION_KWARGS = [
    'name',
    'type',
    'valid'
]


class AccountClient(DomoAPIClient):
    """
        accounts
        - Programmatically manage Domo accounts
        - Docs: https://developer.domo.com/docs/accounts-api-reference/account-api-reference
    """

    def create(self, **kwargs):
        """Create a new account.

        >>> account = { 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        >>> new_account = domo.accounts.create(**account)
        >>> print(new_account)
        {'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}


        :Returns:
          - A dict representing the account
        """

        self._validate_params(kwargs, ACCOUNT_CREATION_KWARGS)

        account_request = kwargs
        print(account_request)
        return self._create(URL_BASE, account_request, {}, ACCOUNT_DESC)

    def get(self, account_id):
        """Get a account.

        >>> account = domo.accounts.get(account_id)
        >>> print(account)
        {'id': '40', 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}

        :Parameters:
          - `account_id`: ID of the account to get (str)

        :returns:
          - A dict representing the account
        """
        url = '{base}/{account_id}'.format(base=URL_BASE, account_id=account_id)
        return self._get(url, ACCOUNT_DESC)

    def list(self, per_account=50, offset=0, limit=0):
        """List accounts.
        Returns a generator that will call the API multiple times
        If limit is supplied and non-zero, returns up to limit accounts

        >>> list(domo.accounts.list())
        [{'id': '40', 'name': 'DataSet Copy Test', ...},
        {'id': '41', 'name': 'DataSet Copy Test2', ...}]

        :Parameters:
        - `per_page`:   results per page. Default 50 (int)
        - `offset`:     offset if you need to paginate results. Default 0 (int)
        - `limit`:      max ouput to return. If 0 then return all results on page. Default 0 (int)
            

        :returns:
          - A list of dicts (with nesting possible)
        """
        # API uses pagination with a max of 50 per account
        if per_account not in range(1, 51):
            raise ValueError('per_account must be between 1 and 50 (inclusive)')
        # Don't pull 50 values if user requests 10
        if limit:
            per_account = min(per_account, limit)

        params = {
            'limit': per_account,
            'offset': offset,
        }
        account_count = 0

        accounts = self._list(URL_BASE, params, ACCOUNT_DESC)
        while accounts:
            for account in accounts:
                yield account
                account_count += 1
                if limit and account_count >= limit:
                    return

            params['offset'] += per_account
            if limit and params['offset'] + per_account > limit:
                # Don't need to pull more than the limit
                params['limit'] = limit - params['offset']
            accounts = self._list(URL_BASE, params, ACCOUNT_DESC)

    def update(self, account_id=None, **kwargs):
        """Update a account.

        >>> print(domo.accounts.get(account_id))
        {'id': '40', 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        updatedAccount = {'name': 'DataSet Copy Test2, 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        >>> domo.accounts.update(account_id, **updatedAccount)
        >>> print(domo.accounts.get(account_id))
        {'id': '40', 'name': 'DataSet Copy Test2, 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}


        :Parameters:
          - `account_id`: ID of the account to update.
          - `kwargs`: New account object
        """

        url = '{base}/{account_id}'.format(base=URL_BASE, account_id=account_id)
        return self._update(url,
                            HTTPMethod.PATCH,
                            requests.codes.accepted,
                            kwargs,
                            ACCOUNT_DESC)

    def delete(self, account_id):
        """Delete a account.

        :Parameters:
          - `account_id`: ID of the account to delete
        """
        url = '{base}/{account_id}'.format(base=URL_BASE, account_id=account_id)
        return self._delete(url, ACCOUNT_DESC)



class DomoObject(dict):
    '''Superclass for DomoAPI objects. These are dict-like objects with
    predefined allowed fields.
    '''
    def __getattribute__(self, name):
        if name in super().__getattribute__('accepted_attrs'):
            return self.__getitem__(name)
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        self.__setitem__(name, value)

    def __setitem__(self, name, value):
        if name not in self.accepted_attrs:
            raise AttributeError("'{}' object has no attribute '{}'".format(
                                                    type(self).__name__, name))
        super().__setitem__(name, value)



from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod
import requests


"""
    Group Client
    - Programmatically manage Domo User Groups
    - Docs: https://developer.domo.com/docs/domo-apis/group-apis
"""


class GroupClient(DomoAPIClient):
    def __init__(self, transport, logger):
        super(GroupClient, self).__init__(transport, logger)
        self.urlBase = '/v1/groups/'
        self.groupDesc = "Group"

    """
        Create a Group
    """
    def create(self, group_request):
        return self._create(self.urlBase, group_request, {}, self.groupDesc)

    """
        Get a Group
    """
    def get(self, group_id):
        return self._get(self._base(group_id), self.groupDesc)

    """
        List Groups
    """
    def list(self, limit, offset):
        params = {
            'limit': str(limit),
            'offset': str(offset),
        }
        return self._list(self.urlBase, params, self.groupDesc)

    """
        Update a Group
    """
    def update(self, group_id, group_update):
        return self._update(self._base(group_id), HTTPMethod.PUT, requests.codes.ok, group_update, self.groupDesc)

    """
        Delete a Group
    """
    def delete(self, group_id):
        return self._delete(self._base(group_id), self.groupDesc)

    """
        Add a User to a Group
    """
    def add_user(self, group_id, user_id):
        url = self._base(group_id) + '/users/' + str(user_id)
        desc = "a User in a Group"
        return self._update(url, HTTPMethod.PUT, requests.codes.no_content, {}, desc)

    """
        List Users in a Group
    """
    def list_users(self, group_id, limit, offset):
        url = self._base(group_id) + '/users'
        desc = "a list of Users in a Group"
        params = {
            'limit': str(limit),
            'offset': str(offset),
        }
        return self._list(url, params, desc)

    """
        Remove a User to a Group
    """
    def remove_user(self, group_id, user_id):
        url = self._base(group_id) + '/users/' + str(user_id)
        desc = "a User in a Group"
        return self._delete(url, desc)



from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod
import pandas as pd
import requests

"""
    User Client
    - Programmatically manage Domo users
    - Docs: https://developer.domo.com/docs/domo-apis/users
"""


class UserClient(DomoAPIClient):
    def __init__(self, transport, logger):
        super(UserClient, self).__init__(transport, logger)
        self.urlBase = '/v1/users/'
        self.userDesc = "User"

    """
        Create a User
    """
    def create(self, user_request, send_invite):
        params = {'sendInvite': str(send_invite)}
        return self._create(self.urlBase, user_request, params, self.userDesc)

    """
        Get a User
    """
    def get(self, user_id):
        return self._get(self._base(user_id), self.userDesc)

    """
        List Users
    """
    def list(self, limit, offset):
        params = {
            'limit': str(limit),
            'offset': str(offset),
        }
        return self._list(self.urlBase, params, self.userDesc)

    """
        Update a User
    """
    def update(self, user_id, user_update):
        return self._update(self._base(user_id), HTTPMethod.PUT, requests.codes.ok, user_update, self.userDesc)

    """
        Delete a User
    """
    def delete(self, user_id):
        return self._delete(self._base(user_id), self.userDesc)

    """
        List all users
    """
    def list_all(self, df_output=True,batch_size=500):
        i = 0
        n_ret = 1
        while n_ret > 0:
            ll = self.list(batch_size,batch_size*i)
            if( i == 0 ):
                all = ll
            else:
                all.extend(ll)
            i = i + 1
            n_ret = ll.__len__()

        out = all

        if( df_output ):
            out = pd.DataFrame(all)

        return(out)



import os
import requests
from pandas import read_csv
from pandas import DataFrame
from io import StringIO

from pydomo.datasets import Sorting, UpdateMethod
from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod

"""
    DataSets
    - Programmatically manage Domo DataSets
    - Use DataSets for fairly static data sources that only require occasional updates via data replacement
    - Use Streams if your data source is massive, constantly changing, or rapidly growing
    - Docs: https://developer.domo.com/docs/data-apis/data
"""

DATA_SET_DESC = "DataSet"
PDP_DESC = "Personalized Data Policy (PDP)"
URL_BASE = '/v1/datasets'


class DataSetClient(DomoAPIClient):
    def __init__(self, transport, logger):
        super(DataSetClient, self).__init__(transport, logger)

    """
        Create a DataSet
    """
    def create(self, dataset_request):
        return self._create(URL_BASE, dataset_request, {}, DATA_SET_DESC)

    """
        Get a DataSet
    """
    def get(self, dataset_id):
        url = '{base}/{dataset_id}'.format(
                base=URL_BASE, dataset_id=dataset_id)
        return self._get(url, DATA_SET_DESC)

    """
        List DataSets
        Returns a generator that will call the API multiple times
        If limit is supplied and non-zero, returns up to limit datasets
    """
    def list(self, sort=Sorting.DEFAULT, per_page=50, offset=0, limit=0):
        # API uses pagination with a max of 50 per page
        if per_page not in range(1, 51):
            raise ValueError('per_page must be between 1 and 50 (inclusive)')

        # Don't pull 50 values if user requests 10
        if limit:
            per_page = min(per_page, limit)

        params = {
            'sort': sort,
            'limit': per_page,
            'offset': offset,
        }
        dataset_count = 0

        datasets = self._list(URL_BASE, params, DATA_SET_DESC)
        while datasets:
            for dataset in datasets:
                yield dataset
                dataset_count += 1
                if limit and dataset_count >= limit:
                    return

            params['offset'] += per_page
            if limit and params['offset'] + per_page > limit:
                # Don't need to pull more than the limit
                params['limit'] = limit - params['offset']
            datasets = self._list(URL_BASE, params, DATA_SET_DESC)

    """
        Update a DataSet
    """
    def update(self, dataset_id, dataset_update):
        url = '{base}/{dataset_id}'.format(
                base=URL_BASE, dataset_id=dataset_id)
        return self._update(url, HTTPMethod.PUT, requests.codes.ok,
                            dataset_update, DATA_SET_DESC)

    """
        Import data from a CSV string
    """
    def data_import(self, dataset_id, csv, update_method=UpdateMethod.REPLACE):
        return self._data_import(dataset_id, str.encode(csv), update_method)

    """
        Import data from a CSV file
    """
    def data_import_from_file(self, dataset_id, filepath,
                              update_method=UpdateMethod.REPLACE):
        with open(os.path.expanduser(filepath), 'rb') as csvfile:
            # passing an open file to the requests library invokes http
            # streaming (uses minimal system memory)
            self._data_import(dataset_id, csvfile, update_method)

    def _data_import(self, dataset_id, csv, update_method):
        url = '{base}/{dataset_id}/data?updateMethod={method}'.format(
                base=URL_BASE, dataset_id=dataset_id, method=update_method)
        return self._upload_csv(url, requests.codes.no_content, csv,
                                DATA_SET_DESC)

    """
        Export data to a CSV string (in-memory)
    """
    def data_export(self, dataset_id, include_csv_header):
        url = '{base}/{dataset_id}/data'.format(
                base=URL_BASE, dataset_id=dataset_id)
        response = self._download_csv(url, include_csv_header)
        if response.status_code == requests.codes.ok:
            return bytes.decode(response.content)
        else:
            self.logger.debug("Error downloading data from DataSet: " + self.transport.dump_response(response))
            raise Exception("Error downloading data from DataSet: " + response.text)

    """
        Export data to a CSV file (streams to disk)
    """
    def data_export_to_file(self, dataset_id, file_path, include_csv_header):
        url = '{base}/{dataset_id}/data'.format(
            base=URL_BASE, dataset_id=dataset_id)
        response = self._download_csv(url, include_csv_header)
        file_path = str(file_path)
        if not file_path.endswith('.csv'):
            file_path += '.csv'
        with open(file_path, 'wb') as csv_file:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    csv_file.write(chunk)
        return open(file_path, 'r+')  # return the file object as readable and writable

    """
        Delete a DataSet
    """
    def delete(self, dataset_id):
        url = '{base}/{dataset_id}'.format(
                base=URL_BASE, dataset_id=dataset_id)
        return self._delete(url, DATA_SET_DESC)

    """
        Create a Personalized Data Policy (PDP)
    """
    def create_pdp(self, dataset_id, pdp_request):
        url = '{base}/{dataset_id}/policies'.format(
                base=URL_BASE, dataset_id=dataset_id)
        return self._create(url, pdp_request, {}, PDP_DESC)

    """
        Get a specific Personalized Data Policy (PDP) for a given DataSet
    """
    def get_pdp(self, dataset_id, policy_id):
        url = '{base}/{dataset_id}/policies/{policy_id}'.format(
                base=URL_BASE, dataset_id=dataset_id, policy_id=policy_id)
        return self._get(url, PDP_DESC)

    """
        List all Personalized Data Policies (PDPs) for a given DataSet
    """
    def list_pdps(self, dataset_id):
        url = '{base}/{dataset_id}/policies'.format(
                base=URL_BASE, dataset_id=dataset_id)
        return self._list(url, {}, DATA_SET_DESC)

    """
        Update a specific Personalized Data Policy (PDP) for a given DataSet
    """
    def update_pdp(self, dataset_id, policy_id, policy_update):
        url = '{base}/{dataset_id}/policies/{policy_id}'.format(
                base=URL_BASE, dataset_id=dataset_id, policy_id=policy_id)
        return self._update(url, HTTPMethod.PUT, requests.codes.ok,
                            policy_update, PDP_DESC)

    """
        Delete a specific Personalized Data Policy (PDPs) for a given DataSet
    """
    def delete_pdp(self, dataset_id, policy_id):
        url = '{base}/{dataset_id}/policies/{policy_id}'.format(
                base=URL_BASE, dataset_id=dataset_id, policy_id=policy_id)
        return self._delete(url, PDP_DESC)
    
    """
        Query's Dataset
    """
    def query(self, dataset_id, query):
        url = '{base}/query/execute/{dataset_id}'.format(
                base=URL_BASE, dataset_id=dataset_id)
        req_body = {'sql':query}
        return self._create(url, req_body, {}, 'query')


import io
import os
import requests

from pydomo.DomoAPIClient import DomoAPIClient
from pydomo.Transport import HTTPMethod

"""
    Streams
    - Programmatically manage Domo Streams
    - A Domo Stream is a specialized upload pipeline pointing to a single Domo DataSet
    - Use Streams for massive, constantly changing, or rapidly growing data sources
    - Streams support uploading data sources in parts, in parallel
    - Use plain DataSets for data sources that only require occasional updates via replacement
    - Usage: Create a Stream, create an Execution, upload data via the Execution, then 'commit' the Execution
    - Docs: https://developer.domo.com/docs/data-apis/data
"""


class StreamClient(DomoAPIClient):
    def __init__(self, transport, logger):
        super(StreamClient, self).__init__(transport, logger)
        self.urlBase = '/v1/streams/'
        self.streamDesc = "Stream"
        self.executionDesc = "Execution"

    """
        Create a Stream and DataSet
        - A Stream is an upload pipeline attached to the newly created DataSet
        - Currently, Streams cannot be applied to existing DataSets
        - Deleting a Stream does not delete the associated DataSet
    """
    def create(self, stream_request):
        return self._create(self.urlBase, stream_request, {}, self.streamDesc)

    """
        Get a Stream
    """
    def get(self, stream_id):
        return self._get(self._base(stream_id), self.streamDesc)

    """
        List Streams
    """
    def list(self, limit, offset):
        params = {
            'limit': str(limit),
            'offset': str(offset),
            'fields': 'all'
        }
        return self._list(self.urlBase, params, self.streamDesc)

    """
        Update a Stream's metadata
    """
    def update(self, stream_id, stream_update):
        return self._update(self._base(stream_id), HTTPMethod.PATCH, requests.codes.ok, stream_update, self.streamDesc)

    """
        Search for Streams by property
    """
    def search(self, stream_property):
        params = {
            'q': str(stream_property),
            'fields': 'all'
        }
        return self._list(self.urlBase + 'search', params, self.streamDesc)

    """
        Create a Stream Execution (begin a multi-part upload process for a single Domo DataSet)
        - A Stream Execution is an upload pipeline that supports uploading data in chunks, or parts
        - Create a new Execution each time you would like to update your Domo DataSet
        - Be sure to 'commit' the Execution once all data parts have been uploaded
        - Do not create multiple Executions on a single Stream

        update_method - update method to use for this execution. If None,
                        the execution will use the stream's current setting.
    """
    def create_execution(self, stream_id, update_method=None):
        url = self._base(stream_id) + '/executions'
        body = {'updateMethod': update_method}
        return self._create(url, body, {}, self.executionDesc)

    """
        Get a Stream Execution
    """
    def get_execution(self, stream_id, execution_id):
        url = self._base(stream_id) + '/executions/' + str(execution_id)
        return self._get(url, self.streamDesc)

    """
        List Stream Executions
    """
    def list_executions(self, stream_id, limit, offset):
        url = self._base(stream_id) + '/executions'
        params = {
            'limit': str(limit),
            'offset': str(offset)
        }
        return self._list(url, params, self.executionDesc)

    """
        Upload a data part (String or CSV)
        - Data sources should be broken into parts and uploaded in parallel
        - Parts should be around 50MB
        - Parts can file-like objects
        - Parts can be compressed
    """
    def upload_part(self, stream_id, execution_id, part_num, csv):
        url = self._base(stream_id) + '/executions/' + str(execution_id) + '/part/' + str(part_num)
        desc = "Data Part on Execution " + str(execution_id) + " on Stream " + str(stream_id)
        if not isinstance(csv, io.IOBase):
            csv = str.encode(csv)
        return self._upload_csv(url, requests.codes.ok, csv, desc)

    """
        Upload a data part CSV file
        - Data sources should be broken into parts and uploaded in parallel
        - Parts should be around 50MB
        - Parts can file-like objects
        - Parts can be compressed
    """
    def upload_csv_part_from_file(self, stream_id, execution_id, part_num, filepath, compression):

        url = self._base(stream_id) + '/executions/' + str(execution_id) + '/part/' + str(part_num)
        desc = "Data Part on Execution " + str(execution_id) + " on Stream " + str(stream_id)

        if compression == 'gzip':

            import gzip

            if filepath.endswith('.gz'):
                with gzip.open(os.path.expanduser(filepath), 'rb') as gzipfile:
                    return self._upload_csv(url, requests.codes.ok, gzipfile, desc)

            else:
                raise ValueError("Valid gzip extension is '.gz'")

        else:
            with open(os.path.expanduser(filepath), 'rb') as csvfile:
                return self._upload_csv(url, requests.codes.ok, csvfile, desc)

    """
        Upload a data part GZIP file
        - Data sources should be broken into parts and uploaded in parallel
        - Parts should be around 50MB
        - Parts can file-like objects
        - Parts can be compressed
    """
    def upload_gzip_part_from_file(self, stream_id, execution_id, part_num, filepath, stream_file, chunk_size):
        
        import gzip

        url = self._base(stream_id) + '/executions/' + str(execution_id) + '/part/' + str(part_num)
        desc = "Data Part on Execution " + str(execution_id) + " on Stream " + str(stream_id)

        if stream_file:
            
            import io

            compressed_body = io.BytesIO()
            compressed_body.name = url
            compressor = gzip.open(compressed_body, mode='wb')

            with open(filepath, 'rb') as csvfile:
                while True:
                    chunk = csvfile.read(chunk_size)
                    if not chunk:
                        break
                    compressor.write(chunk)

            compressor.flush()
            compressor.close()
            compressed_body.seek(0, 0)
            return self._upload_gzip(url, requests.codes.ok, compressed_body, desc)

        else:

            with open(os.path.expanduser(filepath), 'rb') as csvfile:
                compressed_body = gzip.compress(csvfile.read())
                return self._upload_gzip(url, requests.codes.ok, compressed_body, desc)


    """
        Commit an Execution (finalize a multi-part upload process)
        - Finalize a multi-part upload process by committing the execution
        - The execution/upload/commit process is NOT atomic; this is a RESTful approach to multi-part uploading
    """
    def commit_execution(self, stream_id, execution_id):
        url = self._base(stream_id) + '/executions/' + str(execution_id) + '/commit'
        return self._update(url, HTTPMethod.PUT, requests.codes.ok, {}, self.executionDesc)

    """
        Abort an Execution on a given Stream
    """
    def abort_execution(self, stream_id, execution_id):
        url = self._base(stream_id) + '/executions/' + str(execution_id) + '/abort'
        return self._update(url, HTTPMethod.PUT, requests.codes.ok, {}, self.executionDesc)
    """
        Abort the current Execution on a given Stream
    """
    def abort_current_execution(self, stream_id):
        url = self._base(stream_id) + '/executions/abort'
        return self._update(url, HTTPMethod.PUT, requests.codes.no_content, {}, self.executionDesc)

    """
        Delete a Stream
        - Deleting a Stream does not delete the associated DataSet
    """
    def delete(self, stream_id):
        return self._delete(self._base(stream_id), self.streamDesc)



from ..common import DomoObject


class Column(DomoObject):
    accepted_attrs = [
        'type',
        'name'
    ]

    def __init__(self, column_type, name):
        super().__init__()
        self.type = column_type
        self.name = name


class ColumnType:
    STRING = 'STRING'
    DECIMAL = 'DECIMAL'
    LONG = 'LONG'
    DOUBLE = 'DOUBLE'
    DATE = 'DATE'
    DATETIME = 'DATETIME'


class DataSetRequest(DomoObject):
    accepted_attrs = [
        'name',
        'description',
        'schema'
    ]


class FilterOperator:
    EQUALS = 'EQUALS'
    LIKE = 'LIKE'
    GREATER_THAN = 'GREATER_THAN'
    LESS_THAN = 'LESS_THAN'
    GREATER_THAN_EQUAL = 'GREATER_THAN_EQUAL'
    LESS_THAN_EQUAL = 'LESS_THAN_EQUAL'
    BETWEEN = 'BETWEEN'
    BEGINS_WITH = 'BEGINS_WITH'
    ENDS_WITH = 'ENDS_WITH'
    CONTAINS = 'CONTAINS'


class Policy(DomoObject):
    accepted_attrs = [
        'id',
        'type',
        'name',
        'filters',
        'users',
        'groups'
    ]


class PolicyType:
    USER = 'user'
    SYSTEM = 'system'


class PolicyFilter(DomoObject):
    accepted_attrs = [
        'column',
        'values',
        'operator',
        'not'
    ]


class Schema(DomoObject):
    accepted_attrs = [
        'columns'
    ]
    def __init__(self, columns):
        self.columns = columns


class Sorting:
    CARD_COUNT = 'cardCount'
    DEFAULT = None
    NAME = 'name'
    STATUS = 'errorState'
    UPDATED = 'lastUpdated'


class UpdateMethod:
    APPEND = 'APPEND'
    REPLACE = 'REPLACE'



from pydomo.Transport import DomoAPITransport
from pydomo.datasets import DataSetClient
from pydomo.datasets import DataSetRequest
from pydomo.datasets import Schema
from pydomo.datasets import Column
from pydomo.datasets import ColumnType
from pydomo.groups import GroupClient
from pydomo.pages import PageClient
from pydomo.streams import StreamClient
from pydomo.users import UserClient
from pydomo.users import CreateUserRequest
from pydomo.accounts import AccountClient
from pydomo.utilities import UtilitiesClient
from pandas import read_csv
from pandas import DataFrame
from pandas import to_datetime
from io import StringIO
import logging

DOMO = """####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
             ``.############.`          `.###### `#######################``######.`         ``.#####
                 .#######.`                `.###   .###################`   ###.`                `###
    #######..`     .####`     ..######.`     `##     .###############`     ##`     ..######.`     `#
    ###########`    .##     .############.    `#       .###########`       #     .############.
    ############.    #`   `################    `         .#######.         `   `################
    #############`        #################.        ``     .###.     ``        #################.
    #############`        ##################        .##`     `     `##`        ##################
    #############`        #################.        .####`       `####`        #################.
    ############.    #`   `################    `    .######`   `######`    `   `################
    ###########`    .##     .############.     #    .########`########`    #     .############.
    #######..`     .####`     ########.`     `##    .#################`    ##`     ..######.`     `#
                 .#######.`                `.###    .#################`    ###.`                `.##
               .############.`          `.######    .#################`    ######.`          `.#####
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################"""

parent_logger = logging.getLogger('pydomo')
parent_logger.setLevel(logging.WARNING)


class Domo:
    def __init__(self, client_id, client_secret, api_host='api.domo.com', **kwargs):
        if 'logger_name' in kwargs:
            self.logger = parent_logger.getChild(kwargs['logger_name'])
        else:
            self.logger = parent_logger

        if kwargs.get('log_level'):
            self.logger.setLevel(kwargs['log_level'])
        self.logger.debug("\n" + DOMO + "\n")

        self.transport = DomoAPITransport(client_id, client_secret, api_host, kwargs.get('use_https', True), self.logger)
        self.datasets = DataSetClient(self.transport, self.logger)
        self.groups = GroupClient(self.transport, self.logger)
        self.pages = PageClient(self.transport, self.logger)
        self.streams = StreamClient(self.transport, self.logger)
        self.users = UserClient(self.transport, self.logger)
        self.accounts = AccountClient(self.transport, self.logger)
        self.utilities = UtilitiesClient(self.transport, self.logger)

######### Datasets #########
    def ds_meta(self, dataset_id):
        """
            Get a DataSet metadata

            :Parameters:
            - `dataset_id`: id of a dataset (str)

            :Returns:
            - A dict representing the dataset meta-data
        """
        return self.datasets.get(dataset_id)

    def ds_delete(self, dataset_id, prompt_before_delete=True):
        """
            Delete a DataSet naming convention equivalent with rdomo
            
            :Parameters:
            - `dataset_id`: id of a dataset (str)
        """

        del_data = 'Y'
        if prompt_before_delete:
            del_data = input("Permanently delete this data set? This is destructive and cannot be reversed. (Y/n)")

        out = 'Data set not deleted'
        if del_data == 'Y':
            out = self.datasets.delete(dataset_id)

        return out

    def ds_list(self, df_output = True, per_page=50, offset=0, limit=0):
        """
            List DataSets

            >>> l = domo.ds_list(df_output=True)
            >>> print(l.head())

            :Parameters:
            - `df_output`:  should the result be a dataframe. Default True (Boolean)
            - `per_page`:   results per page. Default 50 (int)
            - `offset`:     offset if you need to paginate results. Default 0 (int)
            - `limit`:      max ouput to return. If 0 then return all results on page. Default 0 (int)
            
            :Returns:
            list or pandas dataframe depending on parameters

        """
        l = self.datasets.list()
        if df_output == False:
            out = list(l)
        else:
            out = DataFrame(list(l))
        return out
    
    def ds_query(self, dataset_id, query, return_data=True):
        """
            Evaluate query and return dataset in a dataframe

            >>> query = {"sql": "SELECT * FROM table LIMIT 2"}
            >>> ds = domo.ds_query('80268aef-e6a1-44f6-a84c-f849d9db05fb', query)
            >>> print(ds.head())

            :Parameters:
            - `dataset_id`:     id of a dataset (str)
            - `query`:          query object (dict)
            - `return_data`:    should the result be a dataframe. Default True (Boolean)
            
            :Returns:
            dict or pandas dataframe depending on parameters
        """
        output = self.datasets.query(dataset_id, query)
        if(return_data == True):
            output = DataFrame(output['rows'], columns = output['columns'])
        return output


    def ds_get(self, dataset_id):
        """
            Export data to pandas Dataframe

            >>> df = domo.ds_get('80268aef-e6a1-44f6-a84c-f849d9db05fb')
            >>> print(df.head())

            :Parameters:
            - `dataset_id`:     id of a dataset (str)
            
            :Returns:
            pandas dataframe
        """    
        csv_download = self.datasets.data_export(dataset_id, include_csv_header=True)

        content = StringIO(csv_download)
        df = read_csv(content)

        # Convert to dates or datetimes if possible
        for col in df.columns:
            if df[col].dtype == 'object':
                try:
                    df[col] = to_datetime(df[col])
                except ValueError:
                    pass

        return df

    def ds_create(self, df_up, name, description=''):
        dsr = DataSetRequest()
        dsr.name = name
        dsr.description = description
        dsr.schema = Schema([Column(ColumnType.STRING, 'tt1'),
                             Column(ColumnType.STRING, 'tt2')])

        new_ds_info = self.datasets.create(dsr)

        self.utilities.stream_upload(new_ds_info['id'],df_up,warn_schema_change=False)

        return new_ds_info['id']


    def ds_update(self, ds_id, df_up):
        return self.utilities.stream_upload(ds_id, df_up)

######### PDP #########

    def pdp_create(self, dataset_id, pdp_request):
        """
        Create a PDP policy

        >>> policy = {
          "name": "Only Show Attendees",
          "filters": [ {
            "column": "Attending",
            "values": [ "TRUE" ],
            "operator": "EQUALS"
          } ],
          "users": [ 27 ]
        }
        >>> domo.pdp_create('4405ff58-1957-45f0-82bd-914d989a3ea3', policy)
        {"id" : 8, "type": "user", "name": "Only Show Attendees"
        , "filters": [{"column": "Attending", "values": [ "TRUE" ],   "operator": "EQUALS"
        ,   "not": false } ], "users": [ 27 ],"groups": [ ]}

        :Parameters:
          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required 
          Policy Object:
          - `name`: Name of the Policy (String) Required
          - `filters[].column`:	Name of the column to filter on (String) Required 
          - `filters[].not`: Determines if NOT is applied to the filter operation (Boolean) Required
          - `filters[].operator`: Matching operator (EQUALS) (String) Required 
          - `filters[].values[]`: Values to filter on (String) Required 
          - `type`: Type of policy (user or system) (String) Required 
          - `users`: List of user IDs the policy applies to (array) Required
          - `groups`: List of group IDs the policy applies to (array) Required
        """
        return self.datasets.create_pdp(dataset_id, pdp_request)

    def pdp_delete(self, dataset_id, policy_id):
        """
        Delete PDP Policy

        >>> domo.pdp_delete('4405ff58-1957-45f0-82bd-914d989a3ea3', 35)

        :Parameters:
        - `dataset_id`: id of the dataset PDP will be applied to (String) Required 
        - `policy_id`:  id of the policy to delete (String) Required
        """
        return self.datasets.delete_pdp(dataset_id, policy_id)

    def pdp_list(self, dataset_id, df_output = True):
        """
            List PDP policies

            >>> l = domo.pdp_list(df_output=True)
            >>> print(l.head())

            :Parameters:
            - `dataset_id`:   id of dataset with PDP policies (str) Required
            - `df_output`:  should the result be a dataframe. Default True (Boolean)
            
            :Returns:
            list or pandas dataframe depending on parameters

        """
        output = self.datasets.list_pdps(dataset_id)
        if(df_output == True):
            output = DataFrame(output)
        return output

    def pdp_update(self, dataset_id, policy_id, policy_update):

        """
        Update a PDP policy

        >>> policy = {
          "name": "Only Show Attendees",
          "filters": [ {
            "column": "Attending",
            "values": [ "TRUE" ],
            "operator": "EQUALS"
          } ],
          "users": [ 27 ]
        }
        >>> domo.pdp_create('4405ff58-1957-45f0-82bd-914d989a3ea3', 4, policy)
        {"id" : 8, "type": "user", "name": "Only Show Attendees"
        , "filters": [{"column": "Attending", "values": [ "TRUE" ],   "operator": "EQUALS"
        ,   "not": false } ], "users": [ 27 ],"groups": [ ]}

        :Parameters:
          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required 
          - `policy_id`:    id of the PDP pollicy that will be updated (String) Required 
          Policy Object:
          - `name`: Name of the Policy (String) Required
          - `filters[].column`:	Name of the column to filter on (String) Required 
          - `filters[].not`: Determines if NOT is applied to the filter operation (Boolean) Required
          - `filters[].operator`: Matching operator (EQUALS) (String) Required 
          - `filters[].values[]`: Values to filter on (String) Required 
          - `type`: Type of policy (user or system) (String) Required 
          - `users`: List of user IDs the policy applies to (array) Required
          - `groups`: List of group IDs the policy applies to (array) Required
        """
        return self.datasets.update_pdp(dataset_id, policy_id, policy_update)


######### Pages #########

    def page_create(self, name, **kwargs):
        """Create a new page.

        >>> page = {'name':'New Page'}
        >>> new_page = domo.pages.create(**page)
        >>> print(new_page)
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}

        :Parameters:
          - `name`: The name of the new page
          - `parentId`: (optional) If present create page as subpage
          - `locked`: (optional) whether to lock the page
          - `cardIds`: (optional) cards to place on the page
          - `visibility`: (optional) dict of userIds and/or groupIds to
            give access to

        :Returns:
          - A dict representing the page
        """
        return self.pages.create(name, **kwargs)


    def page_get(self, page_id):
        """Get a page.

        >>> page = domo.pages.get(page_id)
        >>> print(page)
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}

        :Parameters:
          - `page_id`: ID of the page to get

        :returns:
          - A dict representing the page
        """
        return self.pages.get(page_id)


    def page_delete(self, page_id):
        """Delete a page.

        :Parameters:
          - `page_id`: ID of the page to delete
        """
        return self.pages.delete(page_id)


    def collections_create(self, page_id, title, **kwargs):
        """Create a collection on a page.

        >>> collection = domo.pages.create_collection(page_id,
                                                      'Collection')
        >>> print(collection)
        {'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}

        :Parameters:
          - `page_id`: ID of the page to create a collection on
          - `title`: The title of the collection
          - `description`: (optional) The description of the collection
          - `cardIds`: (optional) cards to place in the collection

        :Returns:
          - A dict representing the collection
        """
        return self.pages.create_collection(page_id, title, **kwargs)


    def page_get_collections(self, page_id):
        """Get a collections of a page

        >>> print(domo.pages.get_collections(page_id))
        [{'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}]

        :Parameters:
          - `page_id`: ID of the page

        :Returns:
          - A list of dicts representing the collections
        """
        return self.pages.get_collections(page_id)


    def collections_update(self, page_id, collection_id=None, **kwargs):
        """Update a collection of a page.

        >>> collections = domo.pages.get_collections(page_id)
        >>> print(collections)
        [{'id': 1234321, 'title': 'Collection', 'description': '',
        'cardIds': []}]
        >>> collection_id = collections[0]['id']
        >>> domo.pages.update_collection(page_id, collection_id,
                                         description='description',
                                         cardIds=[54321, 13579])
        >>> print(domo.pages.get_collections(page_id))
        [{'id': 1234321, 'title': 'Collection',
        'description': 'description', 'cardIds': [54321, 13579]}]

        # Using **kwargs:
        >>> collections = domo.pages.get_collections(page_id)
        >>> collections[0]['description'] = 'Description'
        >>> domo.pages.update_collection(page_id, **collections[0])

        :Parameters:
          - `page_id`: ID of the page the collection is on
          - `collection_id`: ID of the collection. Can also be provided
            by supplying `id` to **kwargs. This allows for calling
            get_collections, updating one of the returned collections,
            then passing it to update_collection.
          - `title`: (optional) update the title
          - `description`: (optional) update the description
          - `cardIds`: (optional) update cards in the collection

        :Returns:
          - A dict representing the collection
        """
        return self.pages.update_collection(page_id, collection_id, **kwargs)


    def collections_delete(self, page_id, collection_id):
        """Delete a collection from a page.

        :Parameters:
          - `page_id`: ID of the page the collection is on
          - `collection_id`: ID of the collection to delete
        """
        return self.pages.delete_collection(page_id, collection_id)


    def page_list(self, per_page=50, offset=0, limit=0):
        """List pages.
        Returns a list of dicts (with nesting possible)
        If limit is supplied and non-zero, returns up to limit pages
        """
        return list(self.pages.list())


    def page_update(self, page_id=None, **kwargs):
        """Update a page.

        >>> print(domo.pages.get(page_id))
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': False, 'ownerId': 12345, 'cardIds': [],
        'visibility': {'userIds': 12345}}
        >>> domo.pages.update(page_id, locked=True,
                              cardIds=[54321, 13579])
        >>> print(domo.pages.get(page_id))
        {'id': 123456789, 'parentId': 0, 'name': 'My Page',
        'locked': True, 'ownerId': 12345, 'cardIds': [54321, 13579],
        'visibility': {'userIds': 12345}}

        # Using **kwargs:
        >>> page = domo.pages.get(page_id)
        >>> page['cardIds'].append(new_card_id)
        >>> domo.pages.update(**page)

        :Parameters:
          - `page_id`: ID of the page to update. Can also be provided
            by supplying `id` to **kwargs. This allows for calling get,
            updating the returned object, then passing it to update.
          - `name`: (optional) rename the page
          - `parentId`: (optional) turn page into subpage, or subpage
            into top-level page if parentId is present and falsey
          - `ownerId`: (optional) change owner of the page
          - `locked`: (optional) lock or unlock the page
          - `collectionIds`: (optional) reorder collections on page
          - `cardIds`: (optional) specify which cards to have on page
          - `visibility`: (optional) change who can see the page
        """
        return self.pages.update(page_id, **kwargs)


######### Groups #########

    def groups_add_users(self, group_id, user_id):
        """
            Add a User to a Group
        """

        if isinstance(user_id,list):
            for x in user_id:
                self.groups.add_user(group_id, x)
        else:
            self.groups.add_user(group_id, user_id)

        return 'success'



    def groups_create(self, group_name, users=-1, active='true'):
        """
            Create a Group
        """
        req_body = {'name':group_name,'active':active}
        grp_created = self.groups.create(req_body)
        if (not isinstance(users,list) and users > 0) or isinstance(users,list):
            self.groups_add_users(grp_created['id'],users)

        return grp_created



    def groups_delete(self, group_id):
        """
            Delete a Group
        """
        existing_users = self.groups_list_users(group_id)
        self.groups_remove_users(group_id,existing_users)
        return self.groups.delete(group_id)



    def groups_get(self, group_id):
        """
            Get a Group Definition
        """
        return self.groups.get(group_id)



    def groups_list(self):
        """
            List all groups in Domo instance in a pandas dataframe.
        """
        grps = []
        n_ret = 1
        off = 0
        batch_size = 500
        while n_ret > 0:
            gg = self.groups.list(batch_size,off*batch_size)
            grps.extend(gg)
            n_ret = gg.__len__()
            off += 1
        return DataFrame(grps)


    def groups_list_users(self, group_id):
        """
            List Users in a Group
        """
        user_list = []
        n_ret = 1
        off = 0
        batch_size=500
        while n_ret > 0:
            i_users = self.groups.list_users(group_id,limit=batch_size,offset=off*batch_size)
            user_list.extend(i_users)
            n_ret = i_users.__len__()
            off += 1

        return user_list



    def groups_remove_users(self, group_id, user_id):
        """
            Remove a User to a Group
        """
        if isinstance(user_id,list):
            for x in user_id:
                self.groups.remove_user(group_id, x)
        else:
            self.groups.remove_user(group_id, user_id)

        return 'success'


######### Accounts #########
    def accounts_list(self):
        """List accounts.
        Returns a generator that will call the API multiple times
        If limit is supplied and non-zero, returns up to limit accounts

        >>> list(domo.accounts.list())
        [{'id': '40', 'name': 'DataSet Copy Test', ...},
        {'id': '41', 'name': 'DataSet Copy Test2', ...}]

        :Parameters:
        - `per_page`:   results per page. Default 50 (int)
        - `offset`:     offset if you need to paginate results. Default 0 (int)
        - `limit`:      max ouput to return. If 0 then return all results on page. Default 0 (int)
            

        :returns:
          - A list of dicts (with nesting possible)
        """
        return list(self.accounts.list())

    def accounts_get(self, account_id):
        """Get a account.

        >>> account = domo.accounts.get(account_id)
        >>> print(account)
        {'id': '40', 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}

        :Parameters:
          - `account_id`: ID of the account to get (str)

        :returns:
          - A dict representing the account
        """
        return self.accounts.get(account_id)

    def accounts_delete(self, account_id):
        """Delete a account.

        :Parameters:
          - `account_id`: ID of the account to delete
        """
        return self.accounts.delete(account_id)

    def accounts_create(self, **kwargs):
        """Create a new account.

        >>> account = { 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        >>> new_account = domo.accounts.create(**account)
        >>> print(new_account)
        {'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}


        :Returns:
          - A dict representing the account
        """
        return self.accounts.create(**kwargs)

    def accounts_update(self, account_id, **kwargs):
        """Update a account.

        >>> print(domo.accounts.get(account_id))
        {'id': '40', 'name': 'DataSet Copy Test', 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        updatedAccount = {'name': 'DataSet Copy Test2, 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}
        >>> domo.accounts.update(account_id, **updatedAccount)
        >>> print(domo.accounts.get(account_id))
        {'id': '40', 'name': 'DataSet Copy Test2, 'valid': True, 'type': {'id': 'domo-csv', 'properties': {}}}


        :Parameters:
          - `account_id`: ID of the account to update.
          - `kwargs`: New account object
        """
        return self.accounts.update(account_id, **kwargs)

######### Users #########
    def users_add(self, x_name, x_email, x_role, x_sendInvite=False):
        uu = CreateUserRequest()
        uu.name = x_name
        uu.email = x_email
        uu.role = x_role
        return self.users.create(uu,x_sendInvite)

    def users_get(self, user_id):
        return self.users.get(user_id)

    def users_list(self,df_output=True):
        return self.users.list_all(df_output)

    def users_update(self, user_id, user_def):
        return self.users.update(user_id, user_def)

    def users_delete(self, user_id):
        return self.users.delete(user_id)



from .DataSetModel import Column, ColumnType, DataSetRequest, FilterOperator 
from .DataSetModel import Policy, PolicyType, PolicyFilter, Schema, Sorting
from .DataSetModel import UpdateMethod
from .DataSetClient import DataSetClient



import json
import requests

from pydomo.Transport import HTTPMethod

class DomoAPIClient(object):
    """DomoAPIClient
    Each API client class inherits from this superclass
    """

    def __init__(self, transport, logger):
        self.transport = transport
        self.logger = logger

    def _create(self, url, request, params, obj_desc):
        response = self.transport.post(url=url, params=params, body=request)
        if response.status_code in (requests.codes.CREATED, requests.codes.OK):
            obj = response.json()
            self.print_json("Created", obj)
            return obj
        else:
            self.logger.debug("Error creating " + obj_desc + ": "
                              + self.transport.dump_response(response))
            raise Exception("Error creating " + obj_desc + ": " + response.text)

    def _get(self, url, obj_desc):
        response = self.transport.get(url=url, params={})
        if response.status_code == requests.codes.OK:
            return response.json()
        else:
            self.logger.debug("Error retrieving " + obj_desc + ": "
                              + self.transport.dump_response(response))
            raise Exception("Error retrieving " + obj_desc + ": "
                            + response.text)

    def _update(self, url, method, success_code, obj_update, obj_desc):
        if method == HTTPMethod.PUT:
            response = self.transport.put(url, obj_update)
        elif method == HTTPMethod.PATCH:
            response = self.transport.patch(url, obj_update)
        if response.status_code == success_code:
            if str(response.text) == '':
                return
            else:
                obj = response.json()
                self.print_json("Updated", obj)
                return obj
        else:
            self.logger.debug("Error updating " + obj_desc + ": "
                              + self.transport.dump_response(response))
            raise Exception("Error updating " + obj_desc + ": "
                            + response.text)

    def _list(self, url, params, obj_desc):
        response = self.transport.get(url=url, params=params)
        if response.status_code == requests.codes.OK:
            return response.json()
        else:
            self.logger.debug(obj_desc + " Error: "
                              + self.transport.dump_response(response))
            raise Exception(obj_desc + " Error: " + response.text)

    def _delete(self, url, obj_desc):
        response = self.transport.delete(url=url)
        if response.status_code == requests.codes.NO_CONTENT:
            return
        else:
            self.logger.debug("Error deleting " + obj_desc + ": "
                              + self.transport.dump_response(response))
            raise Exception("Error deleting " + obj_desc + ": "
                            + response.text)

    def _upload_csv(self, url, success_code, csv, obj_desc):
        response = self.transport.put_csv(url=url, body=csv)
        if response.status_code == success_code:
            if str(response.text) == '':
                return
            else:
                return response.json()
        else:
            self.logger.debug("Error uploading " + obj_desc + ": " + self.transport.dump_response(response))
            raise Exception("Error uploading " + obj_desc + ": "
                            + response.text)

    def _upload_gzip(self, url, success_code, csv, obj_desc):
        response = self.transport.put_gzip(url=url, body=csv)
        if response.status_code == success_code:
            if str(response.text) == '':
                return
            else:
                return response.json()
        else:
            self.logger.debug("Error uploading " + obj_desc + ": " + self.transport.dump_response(response))
            raise Exception("Error uploading " + obj_desc + ": "
                            + response.text)

    def _download_csv(self, url, include_csv_header):
        params = {
            'includeHeader': str(include_csv_header)
        }
        return self.transport.get_csv(url=url, params=params)

    def _validate_params(self, params, accepted_keys):
        bad_keys = list(set(params.keys()).difference(accepted_keys))
        if bad_keys:
           raise TypeError('Unexpected keyword arguments: {}'.format(bad_keys))

    def _base(self, obj_id):
        return self.urlBase + str(obj_id)

    def print_json(self, message, json_obj):
        self.logger.info(message + ": " + json.dumps(json_obj, indent=4))



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))
description = 'The official Python3 Domo API SDK - Domo, Inc.'
long_description = 'See https://github.com/domoinc/domo-python-sdk for more details.'

setup(
    name='pydomo',
    version='0.3.0.01',
    description=description,
    long_description=long_description,
    author='Jeremy Morris',
    author_email='jeremy.morris@domo.com',
    url='https://github.com/domoinc/domo-python-sdk',
    download_url='https://github.com/domoinc/domo-python-sdk/tarball/0.2.2.1',
    keywords='domo api sdk',
    license='MIT',
    packages=find_packages(exclude=['examples']),
    install_requires=[
        'requests',
        'requests_toolbelt',
    ],
    python_requires='>=3',
)



from ..common import DomoObject


class CreateStreamRequest(DomoObject):
    accepted_attrs = [
        'dataSet',
        'updateMethod'
    ]
    def __init__(self, data_set_request, update_method):
        super().__init__()
        self.dataSet = data_set_request
        self.updateMethod = update_method


class UpdateMethod:
    APPEND = 'APPEND'
    REPLACE = 'REPLACE'
